--- old/usr/src/uts/common/fs/zfs/dmu_send.c	Tue Nov  6 13:39:29 2012
+++ new/usr/src/uts/common/fs/zfs/dmu_send.c	Tue Nov  6 13:39:29 2012
@@ -643,6 +643,21 @@
 			return (ENODEV);
 	}
 
+	/*
+	 * Check filesystem and snapshot limits before receiving. We'll recheck
+	 * again at the end, but might as well abort before receiving if we're
+	 * already over the limit.
+	 */
+	if (dd->dd_parent != NULL) {
+		err = dsl_dir_fscount_check(dd->dd_parent, 1, NULL);
+		if (err != 0)
+			return (err);
+	}
+
+	err = dsl_snapcount_check(dd, 1, NULL);
+	if (err != 0)
+		return (err);
+
 	return (0);
 }
 
@@ -725,6 +740,11 @@
 		/* if full, most recent snapshot must be $ORIGIN */
 		if (ds->ds_phys->ds_prev_snap_txg >= TXG_INITIAL)
 			return (ENODEV);
+
+		/* Check snapshot limit before receiving */
+		err = dsl_snapcount_check(ds->ds_dir, 1, NULL);
+		if (err != 0)
+			return (err);
 	}
 
 	/* temporary clone name must not exist */
@@ -1547,6 +1567,7 @@
 	char *tosnap;
 	uint64_t creation_time;
 	uint64_t toguid;
+	boolean_t is_new;
 };
 
 static int
@@ -1555,7 +1576,20 @@
 	dsl_dataset_t *ds = arg1;
 	struct recvendsyncarg *resa = arg2;
 
-	return (dsl_dataset_snapshot_check(ds, resa->tosnap, tx));
+	if (resa->is_new) {
+		/* re-check the filesystem limit now that recv is complete */
+		dsl_dir_t *dd;
+		int err;
+
+		dd = ds->ds_dir;
+		if (dd->dd_parent != NULL) {
+			err = dsl_dir_fscount_check(dd->dd_parent, 1, NULL);
+			if (err != 0)
+				return (err);
+		}
+	}
+
+	return (dsl_dataset_snapshot_check(ds, resa->tosnap, 1, tx));
 }
 
 static void
@@ -1564,6 +1598,11 @@
 	dsl_dataset_t *ds = arg1;
 	struct recvendsyncarg *resa = arg2;
 
+	if (resa->is_new)
+		/* update the filesystem counts */
+		dsl_dir_fscount_adjust(ds->ds_dir->dd_parent, tx, 1, B_FALSE,
+		    B_TRUE);
+
 	dsl_dataset_snapshot_sync(ds, resa->tosnap, tx);
 
 	/* set snapshot's creation time and guid */
@@ -1624,6 +1663,7 @@
 	resa.creation_time = drc->drc_drrb->drr_creation_time;
 	resa.toguid = drc->drc_drrb->drr_toguid;
 	resa.tosnap = drc->drc_tosnap;
+	resa.is_new = B_FALSE;
 
 	err = dsl_sync_task_do(ds->ds_dir->dd_pool,
 	    recv_end_check, recv_end_sync, ds, &resa, 3);
@@ -1659,6 +1699,7 @@
 	resa.creation_time = drc->drc_drrb->drr_creation_time;
 	resa.toguid = drc->drc_drrb->drr_toguid;
 	resa.tosnap = drc->drc_tosnap;
+	resa.is_new = B_TRUE;
 
 	err = dsl_sync_task_do(ds->ds_dir->dd_pool,
 	    recv_end_check, recv_end_sync, ds, &resa, 3);
